import os, sys, logging, ConfigParser
###

config_defaults = {
	'general' : {
		'appname' : 'ramona-driven-app',
		'logdir' : '<none>',
		'include' : '<siteconf>',
		'logmaxsize': '{0}'.format(512*1024*1024), # 512Mb
		'logbackups': '3',
	},
	'ramona:server' : {
		'consoleuri': 'tcp://localhost:9876',
		'pidfile': '',
		'log': '<logdir>',
		'loglevel': 'INFO',
	},
	'ramona:console' : {
		'serveruri': 'tcp://localhost:9876',
		'history': '',
	},
	'ramona:smtp' : {
		'smtphost': '',
		'smtpport': '25',
		'sender': '',
	}
}

###

config = ConfigParser.SafeConfigParser()
config.optionxform = str # Disable default 'lowecasing' behavior of ConfigParser
config_files = []
config_includes = []

###

def read_config(configs=None, use_env=True):
	global config
	assert len(config.sections()) == 0

	# Load defaults
	for section, items in config_defaults.iteritems():
		if not config.has_section(section):
			config.add_section(section)

		for key, val in items.iteritems():
			config.set(section, key, val)


	# Load configuration files
	global config_files

	if configs is not None: configs = configs[:]
	else: configs = []
	if use_env:
		# Configs from environment variables
		config_envs = os.environ.get('RAMONA_CONFIG')
		if config_envs is not None:
			for config_file in config_envs.split(':'):
				configs.append(config_file)

	for cfile in  configs:
		if os.path.isfile(cfile):
			config_files.append(cfile)
		config.read([cfile])

	# Handle includes ...
	for _ in range(100):
		includes = config.get('general','include')
		if includes == '': break
		config.set('general','include','')
		includes = includes.split(':')
		for i in xrange(len(includes)-1,-1,-1):
			include = includes[i] = includes[i].strip()
			if include == '<siteconf>':
				# These are platform specific
				siteconfs = ['./site.conf', '/etc/{0}.conf'.format(config.get('general','appname'))]
				includes[i:i+1] = siteconfs
			elif include[:1] == '<':
				L.warning('Unknown include fragment: {0}'.format(include))
				continue

		for include in includes:
			if os.path.isfile(include):
				config_includes.append(include)
				config.read([include])

	else:
		raise RuntimeError("FATAL: It looks like we have loop in configuration includes!")

	# Special treatment of some values
	if config.get('general', 'logdir') == '<none>':
		logdir = os.environ.get('LOGDIR')
		if logdir is None: logdir = '.'
		logdir = os.path.expanduser(logdir)
		config.set('general','logdir',logdir)

###

def get_boolean(value):
	'''
	Translates string/<any-type> value into boolean value. It is kind of similar to ConfigParser.getboolean but this one is used also in different places of code
	'''
	if value is True: return True
	if value is False: return False

	value = str(value)

	if value.upper() in ('TRUE','ON','YES','1'):
		return True
	elif value.upper() in ('FALSE','OFF','NO','0'):
		return False
	else:
		raise ValueError("Invalid boolean string '{0}'' (use one of true, on, yes, false, off or no).".format(value))

###

def get_numeric_loglevel(loglevelstring):
	'''
	Translates log level given in string into numeric value.
	'''
	numeric_level = getattr(logging, loglevelstring.upper(), None)
	if not isinstance(numeric_level, int): raise ValueError('Invalid log level: {0}'.format(loglevelstring))
	return numeric_level



import smtplib, logging, os
from email.mime.text import MIMEText
from .config import config
###

L = logging.getLogger('sendmail')

###

def send_mail(recipient, subject, mail_body):

	smtp_host = config.get('ramona:smtp','host')
	if smtp_host == '':
		L.error('Cannot send mail - SMTP server is not configured.')
		return

	sender = config.get('ramona:smtp','sender')
	if sender is None:
		sender = "{0}@{1}".format(
			# TODO: Does not work on Linux either
			os.getusername(), #TODO: Probably not working in windows - http://stackoverflow.com/questions/842059/is-there-a-portable-way-to-get-the-current-username-in-python
			'xxx'
		)

	msg = MIMEText(mail_body)
	msg['Subject'] = subject
	msg['From'] = sender
	msg['To'] = recipient

	s = smtplib.SMTP( config.getint('ramona:smtp','port'))
	s.sendmail(sender, [recipient], msg.as_string())
	s.quit()



import sys, os, time, logging, shlex, signal, errno, resource
import pyev
from ..config import config, get_boolean
from ..utils import parse_signals, MAXFD, enable_nonblocking, disable_nonblocking
from ..cnscom import program_state_enum
from .logmed import log_mediator

#

L = logging.getLogger("subproc")
Lmy = logging.getLogger("my") # Message yielding logger

#

class program(object):

	DEFAULTS = {
		'command': None,
		'starttimeout': 0.5,
		'stoptimeout': 3,
		'killby': 'TERM,TERM,TERM,QUIT,QUIT,INT,INT,KILL',
		'stdin': '<null>',
		'stdout': '<stderr>',
		'stderr': '<logdir>',
		'priority': 100,
		'disabled': False,
		'coredump': False,
		'autorestart': False,
		'processgroup': True,
	}

	def __init__(self, svrapp, config_section):
		_, self.ident = config_section.split(':', 2)
		self.state = program_state_enum.STOPPED
		self.pid = None

		self.launch_cnt = 0
		self.autorestart_cnt = 0
		self.start_time = None
		self.stop_time = None
		self.exit_time = None
		self.exit_status = None
		self.coredump_enabled = None # If true, kill by SIGQUIT -> dump core

		self.stdout = None
		self.stderr = None
		self.watchers = [
			pyev.Io(0, 0, svrapp.loop, self.__read_stdfd, 0),
			pyev.Io(0, 0, svrapp.loop, self.__read_stdfd, 1),
		]

		# Build configuration
		self.config = self.DEFAULTS.copy()
		self.config.update(config.items(config_section))

		cmd = self.config.get('command')
		if cmd is None:
			L.fatal("Program {0} doesn't specify command - don't know how to launch it".format(self.ident))
			sys.exit(2)

		if cmd == '<httpfend>':
			cmd = '{0} -u -m ramona.httpfend'.format(sys.executable)
		elif cmd[:1] == '<':
			L.error("Unknown command option '{1}' in {0} -> CFGERROR".format(config_section, cmd))
			self.state = program_state_enum.CFGERROR
			return

		self.cmdline = shlex.split(cmd)
		self.stopsignals = parse_signals(self.config['killby'])
		if len(self.stopsignals) == 0: self.stopsignals = [signal.SIGTERM]
		self.act_stopsignals = None

		if self.config['stdin'] != '<null>':
			L.error("Unknown stdin option '{0}' in {1} -> CFGERROR".format(self.config['stdin'], config_section))
			self.state = program_state_enum.CFGERROR
			return

		try:
			self.priority = int(self.config.get('priority'))
		except:
			L.error("Invalid priority option '{0}' in {1} -> CFGERROR".format(self.config['priority'], config_section))
			self.state = program_state_enum.CFGERROR
			return		
		
		try:
			dis = get_boolean(self.config.get('disabled'))
		except ValueError:
			L.error("Unknown 'disabled' option '{0}' in {1} -> CFGERROR".format(dis, config_section))
			self.state = program_state_enum.CFGERROR
			return
		if dis:
			self.state = program_state_enum.DISABLED

		self.ulimits = {}
		#TODO: Enable other ulimits..
		try:
			coredump = get_boolean(self.config.get('coredump',False))
		except ValueError:
			L.error("Unknown 'coredump' option '{0}' in {1} -> CFGERROR".format(self.config.get('coredump','?'), config_section))
			self.state = program_state_enum.CFGERROR
			return
		if coredump: self.ulimits[resource.RLIMIT_CORE] = (-1,-1)

		try:
			self.autorestart = get_boolean(self.config.get('autorestart',False))
		except ValueError:
			L.error("Unknown 'autorestart' option '{0}' in {1} -> CFGERROR".format(self.config.get('autorestart','?'), config_section))
			self.state = program_state_enum.CFGERROR
			return

		try:
			get_boolean(self.config.get('processgroup',True))
		except ValueError:
			L.error("Unknown 'processgroup' option '{0}' in {1} -> CFGERROR".format(self.config.get('processgroup','?'), config_section))
			self.state = program_state_enum.CFGERROR
			return


		# Prepare log files
		stdout_cnf = self.config['stdout']
		stderr_cnf = self.config['stderr']

		if (stdout_cnf == '<stderr>') and (stderr_cnf == '<stdout>'):
			L.error("Invalid stdout and stderr combination in {0} -> CFGERROR".format(config_section))
			self.state = program_state_enum.CFGERROR
			return			

		# Stdout settings
		if stdout_cnf == '<logdir>':
			if stderr_cnf  in ('<stderr>','<null>') :
				fname = os.path.join(config.get('general','logdir'), self.ident + '.log')
			else:
				fname = os.path.join(config.get('general','logdir'), self.ident + '-out.log')
			self.log_out = log_mediator(fname)
		elif stdout_cnf == '<stderr>':
			pass
		elif stdout_cnf == '<null>':
			self.log_out = log_mediator(None)
		elif stdout_cnf[:1] == '<':
			L.error("Unknown stdout option in {0} -> CFGERROR".format(config_section))
			self.state = program_state_enum.CFGERROR
			return			
		else:
			self.log_out = log_mediator(stdout_cnf)

		# Stderr settings
		if stderr_cnf == '<logdir>':
			if stdout_cnf in ('<stderr>','<null>') :
				fname = os.path.join(config.get('general','logdir'), self.ident + '.log')
			else:
				fname = os.path.join(config.get('general','logdir'), self.ident + '-err.log')
			self.log_err = log_mediator(fname)
		elif stderr_cnf == '<stdout>':
			self.log_err = self.log_out
		elif stderr_cnf == '<null>':
			self.log_err = log_mediator(None)
		elif stderr_cnf[:1] == '<':
			L.error("Unknown stderr option in {0} -> CFGERROR".format(config_section))
			self.state = program_state_enum.CFGERROR
			return
		else:
			self.log_err = log_mediator(stderr_cnf)

		if stdout_cnf == '<stderr>':
			self.log_out = self.log_err

		# Environment variables
		self.env = os.environ.copy()
		if config.has_section('env'):
			for name, value in config.items('env'):
				if value != '':
					self.env[name] = value
				else:
					self.env.pop(name, 0)
		self.env['RAMONA_SECTION'] = config_section


	def __repr__(self):
		return "<{0} {1} state={2} pid={3}>".format(self.__class__.__name__, self.ident, program_state_enum.labels[self.state],self.pid if self.pid is not None else '?')


	def spawn(self, cmd, args):
		self.stdout, stdout = os.pipe()
		self.stderr, stderr = os.pipe()

		pid = os.fork()
		if pid !=0:
			os.close(stdout)
			os.close(stderr)

			enable_nonblocking(self.stdout)
			self.watchers[0].set(self.stdout, pyev.EV_READ)
			self.watchers[0].start()

			enable_nonblocking(self.stderr)
			self.watchers[1].set(self.stderr, pyev.EV_READ)
			self.watchers[1].start()

			return pid

		try:
			# Launch in dedicated process group (optionally)
			if get_boolean(self.config.get('processgroup',True)):
				os.setsid()

			# Stdin/stdout/stderr
			if self.config['stdin'] == '<null>':
				stdin = os.open(os.devnull, os.O_RDONLY) # Open stdin
			else:
				# Default is to open /dev/null
				stdin = os.open(os.devnull, os.O_RDONLY) # Open stdin
			os.dup2(stdin, 0)
			os.dup2(stdout, 1) # Prepare stdout
			os.dup2(stderr, 2) # Prepare stderr

			# Close all open file descriptors above standard ones.  This prevents the child from keeping
			# open any file descriptors inherited from the parent.
			os.closerange(3, MAXFD)

			# Set ulimits
			for k,v in self.ulimits.iteritems():
				try:
					resource.setrlimit(k,v)
				except Exception, e:
					os.write(2, "WARNING: Setting ulimit '{1}' failed: {0}\n".format(e, k))

			try:
				os.execvpe(cmd, args, self.env)
			except Exception, e:
				os.close(1)
				os.write(2, "FATAL: Execution of command '{1}' failed: {0}\n".format(e, cmd))
				os.close(2)

		finally:
			# No pasaran
			os._exit(3)


	def start(self, reset_autorestart_cnt=True):
		'''Transition to state STARTING'''
		assert self.state in (program_state_enum.STOPPED, program_state_enum.FATAL)

		L.debug("{0} -> STARTING".format(self))

		self.pid = self.spawn(self.cmdline[0], self.cmdline) #TODO: self.cmdline[0] can be substituted by self.ident or any arbitrary string
		self.log_out.open()
		self.log_err.open()
		self.log_err.write("\n------[ STARTING by Ramona on {0} ]------\n".format(time.strftime("%Y-%m-%d %H:%M:%S")))
		self.state = program_state_enum.STARTING
		self.start_time = time.time()
		self.stop_time = None
		self.exit_time = None
		self.exit_status = None
		self.coredump_enabled = None
		self.launch_cnt += 1
		if reset_autorestart_cnt: self.autorestart_cnt = 0


	def stop(self):
		'''Transition to state STOPPING'''
		if self.state == program_state_enum.FATAL: return # This can happen and it is probably OK

		assert self.pid is not None, "Stopping: {0}".format(self)
		assert self.state in (program_state_enum.RUNNING, program_state_enum.STARTING)

		L.debug("{0} -> STOPPING".format(self))
		self.act_stopsignals = self.stopsignals[:]
		signal = self.get_next_stopsignal()
		try:
			if get_boolean(self.config.get('processgroup',True)):
				os.kill(-self.pid, signal) # Killing whole process group
			else:
				os.kill(self.pid, signal)
		except:
			pass
		self.state = program_state_enum.STOPPING
		self.stop_time = time.time()


	def on_terminate(self, status):
		self.exit_time = time.time()
		self.exit_status = status
		self.pid = None

		# Close process stdout and stderr pipes (including vacuum of actual content)
		self.watchers[0].stop()
		if self.stdout is not None:
			disable_nonblocking(self.stdout)
			while True:
				signal.setitimer(signal.ITIMER_REAL, 0.5) # Set timeout for following operation
				try:
					data = os.read(self.stdout, 4096)
				except OSError, e:
					if e.errno == errno.EINTR:
						L.warning("We have stall recovery situation on stdout socket of {0}".format(self))
						# This stall situation can happen when program shares stdout with its child
						# e.g. command=bash -c "echo ahoj1; tail -f /dev/null"
						break
					raise
				if len(data) == 0: break
				self.log_out.write(data)
			os.close(self.stdout)
			self.stdout = None

		self.watchers[1].stop()
		if self.stderr is not None:
			disable_nonblocking(self.stderr)
			while True:
				signal.setitimer(signal.ITIMER_REAL, 0.2) # Set timeout for following operation
				try:
					data = os.read(self.stderr, 4096)
				except OSError, e:
					if e.errno == errno.EINTR:
						L.warning("We have stall recovery situation on stderr socket of {0}".format(self))
						# See comment above
						break
					raise
				if len(data) == 0: break
				self.log_err.write(data)
			os.close(self.stderr)
			self.stderr = None

		# Close log files
		self.log_err.write("\n------[ EXITED on {0} with status {1} ]------\n".format(time.strftime("%Y-%m-%d %H:%M:%S"), status))
		self.log_out.close()
		self.log_err.close()

		# Handle state change properly
		if self.state == program_state_enum.STARTING:
			Lmy.error("{0} exited too quickly (now in FATAL state)".format(self.ident))
			L.error("{0} exited too quickly -> FATAL".format(self))
			self.state = program_state_enum.FATAL

		elif self.state == program_state_enum.STOPPING:
			Lmy.info("{0} is now STOPPED".format(self.ident))
			L.debug("{0} -> STOPPED".format(self))
			self.state = program_state_enum.STOPPED

		else:
			if self.autorestart:
				Lmy.error("{0} exited unexpectedly and going to be restarted".format(self.ident))
				L.error("{0} exited unexpectedly -> FATAL -> autorestart".format(self))
				self.state = program_state_enum.FATAL
				self.autorestart_cnt += 1
				self.start(reset_autorestart_cnt=False)
			else:
				Lmy.error("{0} exited unexpectedly (now in FATAL state)".format(self.ident))
				L.error("{0} exited unexpectedly -> FATAL".format(self))
				self.state = program_state_enum.FATAL


	def on_tick(self, now):
		# Switch starting programs into running state
		if self.state == program_state_enum.STARTING:
			if now - self.start_time >= self.config['starttimeout']:
				Lmy.info("{0} is now RUNNING".format(self.ident))
				L.debug("{0} -> RUNNING".format(self))
				self.state = program_state_enum.RUNNING

		elif self.state == program_state_enum.STOPPING:
			if now - self.stop_time >= self.config['stoptimeout']:
				L.warning("{0} is still terminating - sending another signal".format(self))
				signal = self.get_next_stopsignal()
				try:
					if get_boolean(self.config.get('processgroup',True)):
						os.kill(-self.pid, signal) # Killing whole process group
					else:
						os.kill(self.pid, signal)
				except:
					pass


	def get_next_stopsignal(self):
		if self.coredump_enabled:
			self.coredump_enabled = None
			return signal.SIGQUIT
		if len(self.act_stopsignals) == 0: return signal.SIGKILL
		return self.act_stopsignals.pop(0)


	def __read_stdfd(self, watcher, revents):
		while 1:
			try:
				data = os.read(watcher.fd, 4096)
			except OSError, e:
				if e.errno == errno.EAGAIN: return # No more data to read (would block)
				raise

			if len(data) == 0: # File descriptor is closed
				watcher.stop()
				os.close(watcher.fd)
				if watcher.data == 0: self.stdout = None
				elif watcher.data == 1: self.stderr = None
				return 
			
			if watcher.data == 0: self.log_out.write(data)
			elif watcher.data == 1: self.log_err.write(data)


	def tail(self, stream):
		if stream == 'stdout':
			return self.log_out.tail()
		elif stream == 'stderr':
			return self.log_err.tail()
		else:
			raise ValueError("Unknown stream '{0}'".format(stream))


	def charge_coredump(self):
		l = self.ulimits.get(resource.RLIMIT_CORE, (0,0))
		if l == (0,0):
			Lmy.warning("Program {0} is not configured to dump code".format(self.ident))
			return
		self.coredump_enabled = True



import json
from ... import cnscom
from .. import exception
###

name = 'start'
cmdhelp = 'Launch subprocess(es)'

###

def init_parser(parser):
	parser.add_argument('-n','--no-server-start', action='store_true', help='Avoid eventual automatic server start')
	parser.add_argument('-i','--immediate-return', action='store_true', help='Dont wait for start of subprocesses and return ASAP')
	parser.add_argument('-f','--force-start', action='store_true', help='Force start of processes in FATAL state')
	parser.add_argument('-S','--server-only', action='store_true', help='Start only server, programs are not launched')
	parser.add_argument('program', nargs='*', help='Optionally specify program(s) in scope of the command')

###

def main(cnsapp, args):
	if args.server_only:
		if len(args.program) > 0:
			raise exception.parameters_error('Cannot specify programs and -S option at once.')
		cnsapp.auto_server_start()
		return

	params={
		'force': args.force_start,
		'immediate': args.immediate_return,
	}
	if len(args.program) > 0: params['pfilter'] = args.program

	cnsapp.svrcall(
		cnscom.callid_start,
		json.dumps(params),
		auto_connect=args.no_server_start,
		auto_server_start=not args.no_server_start
	)




import os, sys, signal, resource, fcntl, logging
###

L = logging.getLogger("utils")

###

def launch_server():
	'''
This function launches Ramona server - in 'os.exec' manner which means that this function will not return
and instead of that, current process will be replaced by launched server. 

All file descriptors above 2 are closed.
	'''
	from .config import config_files

	env = os.environ.copy()
	env['RAMONA_CONFIG'] = ':'.join(config_files)
	
	# Close all open file descriptors above standard ones.  This prevents the child from keeping
	# open any file descriptors inherited from the parent.
	os.closerange(3, MAXFD)

	os.execle(sys.executable, sys.executable, "-m", "ramona.server", env)

#

def launch_server_daemonized():
	"""
This function launches Ramona server as a UNIX daemon.
It detaches the process context from parent (caller) and session.
In comparison to launch_server() it returns.
	"""
	from .config import config

	logfname = config.get('ramona:server','log')
	if logfname == '<logdir>':
		logfname = os.path.join(config.get('general','logdir'), 'ramona.log')
	elif logfname[:1] == '<':
		L.error("Unknown log option in [server] section - server not started")
		return

	try:
		logf = open(logfname, 'a')
	except IOError, e:
		L.fatal("Cannot open logfile {0} for writing: {1}. Check the configuration in [server] section. Exiting.".format(logfname, e))
		return

	with logf:
		pid = os.fork()
		if pid > 0:
			return pid

		os.setsid()

		pid = os.fork()
		if pid > 0:
			os._exit(0)

		stdin = os.open(os.devnull, os.O_RDONLY)
		os.dup2(stdin, 0)

		os.dup2(logf.fileno(), 1) # Prepare stdout
		os.dup2(logf.fileno(), 2) # Prepare stderr
	launch_server()

###

def parse_signals(signals):
	ret = []
	signame2signum = dict((name, num) for name, num in signal.__dict__.iteritems() if name.startswith('SIG') and not name.startswith('SIG_'))
	for signame in signals.split(','):
		signame = signame.strip().upper()
		if not signame.startswith('SIG'): signame = 'SIG'+signame
		signum = signame2signum.get(signame)
		if signum is None: raise RuntimeError("Unknown signal '{0}'".format(signame))
		ret.append(signum)
	return ret

###

MAXFD = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
if (MAXFD == resource.RLIM_INFINITY): MAXFD = 1024

###

def enable_nonblocking(fd):
	fl = fcntl.fcntl(fd, fcntl.F_GETFL)
	fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

def disable_nonblocking(fd):
	fl = fcntl.fcntl(fd, fcntl.F_GETFL)
	fcntl.fcntl(fd, fcntl.F_SETFL, fl ^ os.O_NONBLOCK)




from distutils.core import setup

# See http://guide.python-distribute.org/
# See http://docs.python.org/distutils/setupscript.html

setup(
	name='ramona',
	description='Enterprise-grade runtime supervisor',
	author='Ales Teska',
	author_email='ales.teska+ramona@gmail.com',
	version='0.9b1', # Also in ramona.__init__.py (+ relevant version format specification)
	packages=['ramona','ramona.server','ramona.console','ramona.console.cmd','ramona.httpfend'],
	license='BSD 2-clause "Simplified" License',
	long_description=open('README').read(),
	url='https://github.com/ateska/ramona',
	requires="pyev",
	package_data={
		'ramona.httpfend': [
			'*.html',
			'static/jquery/*.js',
			'static/bootstrap/js/*.js',
			'static/bootstrap/css/*.css',
			'static/bootstrap/img/*.png',
			'static/img/*.gif',
			'static/img/*.ico',
			]
	},
	classifiers=[
		'Development Status :: 4 - Beta',
		'Environment :: Console',
		'Intended Audience :: Developers',
		'Intended Audience :: Information Technology',
		'Intended Audience :: System Administrators',
		'License :: OSI Approved :: BSD License',
		'Natural Language :: English',
		'Operating System :: MacOS :: MacOS X',
		'Operating System :: Microsoft :: Windows',
		'Operating System :: POSIX',
		'Operating System :: Unix',
		'Programming Language :: Python :: 2.7',
		'Topic :: Software Development',
		'Topic :: System :: Monitoring',
		'Topic :: System :: Systems Administration',
	],
)




import socket, errno, struct, weakref, json, select, logging
import pyev
from .. import cnscom
###

L = logging.getLogger("cnscon")

###

class console_connection(object):
	'''Server side of console communication IPC'''

	NONBLOCKING = frozenset([errno.EAGAIN, errno.EWOULDBLOCK])

	def __init__(self, sock, address, serverapp):
		self.serverapp = serverapp

		self.sock = sock
		self.address = address
		self.sock.setblocking(0)
		
		self.read_buf = ""
		self.write_buf = None
		
		self.yield_enabled = False

		self.watcher = pyev.Io(self.sock._sock, pyev.EV_READ, serverapp.loop, self.io_cb)
		self.watcher.start()

		L.debug("Console connection open ({0})".format(self.address))


	def __del__(self):
		self.close()


	def reset(self, events):
		self.watcher.stop()
		self.watcher.set(self.sock, events)
		self.watcher.start()


	def io_cb(self, watcher, revents):
		try:
			if (revents & pyev.EV_READ) == pyev.EV_READ:
				self.handle_read()

			if self.sock is None: return # Socket has been just closed

			if (revents & pyev.EV_WRITE) == pyev.EV_WRITE:
				self.handle_write()
		except:
			L.exception("Exception during IO on console connection:")


	def handle_read(self):
		try:
			buf = self.sock.recv(1024)
		except socket.error as err:
			if err.args[0] not in self.NONBLOCKING:
				L.error("Error when reading from console connection socket: {0}".format(err)) 
				self.handle_error()
			return
		
		if len(buf) > 0:
			self.read_buf += buf
			
			while len(self.read_buf) >= 4:
				magic, callid, paramlen = struct.unpack(cnscom.call_struct_fmt, self.read_buf[:4])
				if magic != cnscom.call_magic:
					L.warning("Invalid data stream on control port")
					self.handle_error()
					return

				if (paramlen + 4) <= len(self.read_buf):
					params = self.read_buf[4:4+paramlen]
					self.read_buf = self.read_buf[4+paramlen:]
					
					try:
						ret = self.serverapp.dispatch_svrcall(self, callid, params)
					except Exception, e:
						if not isinstance(e, cnscom.svrcall_error):
							L.exception("Exception during dispatching console call")
						self.send_exception(e, callid)
					else:
						if ret == deffered_return: return
						self.send_return(ret, callid)

		else:
			L.debug("Connection closed by peer")
			self.handle_error()


	def handle_write(self):
		try:
			sent = self.sock.send(self.write_buf[:select.PIPE_BUF])
		except socket.error as err:
			if err.args[0] not in self.NONBLOCKING:
				#TODO: Log "error writing to {0}".format(self.sock)
				self.handle_error()
				return
		else :
			self.write_buf = self.write_buf[sent:]
			if len(self.write_buf) == 0:
				self.reset(pyev.EV_READ)
				self.write_buf = None


	def write(self, data):
		if self.sock is None:
			L.warning("Socket is closed - write operation is ignored")
			return

		if self.write_buf is None:
			self.write_buf = data
			self.reset(pyev.EV_READ | pyev.EV_WRITE)
		else:
			self.write_buf += data


	def close(self):
		if self.watcher is not None:
			self.watcher.stop()
			self.watcher = None
		if self.sock is not None:
			self.sock.close()
			self.sock = None


	def handle_error(self):
		L.debug("Console connection closed.")
		self.close()


	def send_return(self, ret, callid='-'):
		'''
		Internal function that manages communication of response (type return) to the console (client).
		'''
		self.yield_enabled = False
		ret = str(ret)
		lenret = len(ret)
		if lenret >= 0x7fff:
			self.handle_error()
			raise RuntimeError("Transmitted return value is too long (callid={0})".format(callid))
		self.write(struct.pack(cnscom.resp_struct_fmt, cnscom.resp_magic, cnscom.resp_return, lenret) + ret)


	def send_exception(self, e, callid='-'):
		'''
		Internal function that manages communication of response (type exception) to the console (client).
		'''
		self.yield_enabled = False
		ret = str(e)
		lenret = len(ret)
		if lenret >= 0x7fff:
			self.handle_error()
			raise RuntimeError("Transmitted exception is too long (callid={0})".format(callid))
		self.write(struct.pack(cnscom.resp_struct_fmt, cnscom.resp_magic, cnscom.resp_exception, lenret) + ret)


	def yield_message(self, message):
		if not self.yield_enabled: return

		messagelen = len(message)
		if messagelen >= 0x7fff:
			raise RuntimeError("Transmitted yield message is too long.")

		self.write(struct.pack(cnscom.resp_struct_fmt, cnscom.resp_magic, cnscom.resp_yield_message, messagelen) + message)

###

class message_yield_loghandler(logging.Handler):
	'''
	Message yield(ing) log handler provides functionality to propagate log messages to connected consoles.
	It automatically emits all log records that are submitted into relevant logger (e.g. Lmy = logging.getLogger("my") ) and forwards them
	as resp_yield_message to connected consoles (yield has to be enabled on particular connection see yield_enabled).
	'''


	def __init__(self, serverapp):
		logging.Handler.__init__(self)
		self.serverapp = weakref.ref(serverapp)


	def emit(self, record):
		serverapp = self.serverapp()
		if serverapp is None: return

		msg = json.dumps({
			'msg': record.msg,
			'args': record.args,
			'funcName': record.funcName,
			'lineno': record.lineno,
			'levelno': record.levelno,
			'levelname': record.levelname,
			'name': record.name,
			'pathname': record.pathname,
		})

		for conn in serverapp.conns:
			conn.yield_message(msg)

###

class deffered_return(object): pass # This is just a symbol definition



import os, cmd, logging, sys
from ...config import config
from ... import cnscom

###

name = 'console'
cmdhelp = 'Enter interactive console mode'

###

L = logging.getLogger("console")

###

def init_parser(parser):
	return

###

import readline #TODO: optional

# See http://stackoverflow.com/questions/7116038/python-tab-completion-mac-osx-10-7-lion
if 'libedit' in readline.__doc__:
	readline.parse_and_bind("bind ^I rl_complete")
else:
	readline.parse_and_bind("tab: complete")

###

class _console_cmd(cmd.Cmd):


	def __init__(self, cnsapp):
		self.prompt = '> '
		self.cnsapp = cnsapp

		from ..parser import consoleparser
		self.parser = consoleparser(self.cnsapp)

		# Build dummy method for each command in the parser
		for m in self.parser.subcommands.keys():
			def do_cmd_template(self, _cmdline):
				try:
					self.parser.execute(self.cnsapp)
				except Exception, e:
					L.error("{0}".format(e))

			setattr(self.__class__, "do_{0}".format(m), do_cmd_template)

		cmd.Cmd.__init__(self)
	

	def precmd(self, line):
		if line == '': return ''
		if line == "EOF":
			print
			sys.exit(0)
			
		try:
			self.parser.parse(line.split())
		except SyntaxError: # To capture cases like 'xxx' (avoid exiting)
			self.parser.parse(['help'])
			return 'help'
		except SystemExit: # To capture cases like 'tail -xxx' (avoid exiting)
			self.parser.parse(['help'])
			return 'help'
		return line


	def emptyline(self):
		# Send 'ping' to server
		try:
			self.cnsapp.svrcall(cnscom.callid_ping, '', auto_connect=True)
		except Exception, e:
			L.error("{0}".format(e))
	
#

def main(cnsapp, args):
	old_is_interactive = cnsapp.is_interactive
	cnsapp.is_interactive = True
	try:
		L.info("Ramona console for {0}".format(config.get('general','appname'))) #TODO: Add version info

		histfile = config.get('ramona:console', 'history')
		if histfile != '':
			histfile = os.path.expanduser(histfile)
			try:
				readline.read_history_file(histfile)
			except IOError:
				pass

		c = _console_cmd(cnsapp)
		try:
			c.cmdloop()
		
		except Exception, e:
			L.exception("Exception during cmd loop:")

		except KeyboardInterrupt:
			print ""
		
		finally:
			if histfile != '':
				try:
					readline.write_history_file(histfile)
				except Exception, e:
					L.warning("Cannot write console history file '{1}': {0}".format(e, histfile))
	finally:
		cnsapp.is_interactive = old_is_interactive
